# CHANGELOG
{# RELEASED #}
{% for version, release in context.history.released.items() %}
## {{ version.as_tag() }} ({{ release.tagged_date.strftime("%Y-%m-%d") }})
{% for type_, commits in release["elements"] | dictsort %}
{% if type_ != "unknown" %}
### {{ type_ | capitalize }}
{% for commit in commits %}
* {{ commit.commit.message.split("\n")[0].rstrip()}} ([`{{ commit.commit.hexsha[:7] }}`]({{ commit.commit.hexsha | commit_hash_url }}))
{% endfor %}{% endif %}{% endfor %}{% endfor %}
